#!/usr/bin/env python3
#
# Copyright (c) 2019, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

from collections import defaultdict, OrderedDict

from .format import ColorWrapper as _cw
from .node import BaseNode
from .option import *
import lbuild.utils


class CollectorContext:
    def __init__(self, module, filename=None):
        self.module = module
        self.filename = filename

    @property
    def repository(self):
        return self.module.split(":")[0]

    @property
    def has_filename(self):
        return self.filename is not None


class Collector(BaseNode):

    def __init__(self, option):
        BaseNode.__init__(self, option.name, BaseNode.Type.COLLECTOR)

        self._option = option
        self._description = option._description
        self._values = OrderedDict()

    @property
    def module(self):
        return self.parent

    def add_values(self, values, module, operations=None):
        checked_values = list()
        for value in lbuild.utils.listify(values):
            self._option.value = value
            checked_values.append(self._option.value)

        if operations is not None:
            for operation in operations:
                context = CollectorContext(operation.module, operation.filename)
                self._extend_values(context, checked_values)
        else:
            self._extend_values(CollectorContext(module.fullname), checked_values)

    def _extend_values(self, context, values):
        if context not in self._values:
            self._values[context] = []
        self._values[context].extend(values)

    def values(self, default=None, filterfunc=None, unique=True):
        if filterfunc is None:
            values = [v  for vs in self._values.values()  for v in vs]
        else:
            values = []
            for context, vals in self._values.items():
                if filterfunc(context):
                    values.extend(vals)
        if unique:
            values = list(OrderedDict.fromkeys(values))
        if default is not None and not values:
            values = lbuild.utils.listify(default)
        return values

    def items(self):
        return self._values.items()

    def keys(self):
        return self._values.keys()

    @property
    def class_name(self):
        return self._option.class_name

    def format_values(self):
        return self._option.format_values()


# Inherited Options as Collectors

class StringCollector(StringOption):
    def __init__(self, name, description, validate=None):
        StringOption.__init__(self, name, description, validate=validate)


class PathCollector(PathOption):
    def __init__(self, name, description, empty_ok=False):
        PathOption.__init__(self, name, description, empty_ok=empty_ok)


class BooleanCollector(BooleanOption):
    def __init__(self, name, description):
        BooleanOption.__init__(self, name, description)


class NumericCollector(NumericOption):
    def __init__(self, name, description, minimum=None, maximum=None):
        NumericOption.__init__(self, name, description, minimum, maximum)


class EnumerationCollector(EnumerationOption):
    def __init__(self, name, description, enumeration):
        EnumerationOption.__init__(self, name, description, enumeration)


class CallableCollector(Option):
    def __init__(self, name, description):
        Option.__init__(self, name, description,
                        convert_input=self.as_callable,
                        convert_output=lambda f: f)

    @staticmethod
    def as_callable(func):
        if not callable(func):
            raise ValueError("Object '{}' must be callable!".format(func))
        return func

    @property
    def values(self):
        return ["Callable"]

    def format_values(self):
        return _cw("Callable")


#!/usr/bin/env python3
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import sys
import enum
import os.path
import random
import logging
import collections

import lbuild.module
import lbuild.environment

from .exception import LbuildException
from .exception import LbuildBuildException

from .node import BaseNode
from .config import ConfigNode
from .facade import BuildLogFacade

from . import repository
from . import utils

LOGGER = logging.getLogger('lbuild.parser')


class Runner:

    def __init__(self, node, env):
        self.node = node
        self.env = env

    def validate(self):
        if hasattr(self.node, "validate"):
            self.env.stage = Parser.Stage.VALIDATE
            self.node.validate(self.env)

    def build(self):
        self.env.stage = Parser.Stage.BUILD
        self.node.build(self.env)

    def post_build(self, buildlog):
        if hasattr(self.node, "post_build"):
            self.env.stage = Parser.Stage.POST_BUILD
            self.node.post_build(self.env)


class Parser(BaseNode):

    @enum.unique
    class Stage(enum.IntEnum):
        """This order of repo/module function calls"""
        INIT = 1
        PREPARE = 2
        VALIDATE = 3
        BUILD = 4
        POST_BUILD = 5

    def __init__(self, config=None):
        BaseNode.__init__(self, "lbuild", BaseNode.Type.PARSER)
        self._config = config if config else ConfigNode()
        self._config_flat = self._config.flatten()

    @property
    def config(self):
        return self._config_flat

    @property
    def modules(self):
        return {m.fullname:m for m in self.all_modules()}

    @property
    def repositories(self):
        return {r.fullname:r for r in self._findall(BaseNode.Type.REPOSITORY)}

    @property
    def repo_options(self):
        return {o.fullname:o for o in self.all_options(depth=3)}

    @property
    def module_options(self):
        return {o.fullname:o for o in self.all_options() if o.depth > 2}

    def load_repositories(self, repofilenames=None):
        repofiles = set(os.path.realpath(p) for p in utils.listify(repofilenames))
        parsed = set()
        config_map = {}

        while True:
            # flatten configuration and collect repositories
            self._config_flat = self._config.flatten()
            repofiles = (repofiles | set(self._config_flat.repositories)) - parsed
            if not len(repofiles) + len(parsed):
                LOGGER.error("\n%s", self._config.render())
                raise LbuildException("No repositories loaded!")
            parsed |= repofiles

            # Parse only new repositories
            for repofile in repofiles:
                repo = self.parse_repository(repofile)
                config_map.update({c.fullname:c._config for c in repo.configurations})

            # nothing more to extend
            if not self._config_flat._extends:
                break

            for filename, aliases in self._config_flat._extends.items():
                node = self._config.find(filename)
                for alias in aliases:
                    if alias not in config_map:
                        raise LbuildException("Configuration alias '{}' not found in any map! "
                                              "Available aliases: '{}'"
                                              .format(alias,
                                                      "', '".join(config_map)))
                    self._config.extend(node, ConfigNode.from_file(config_map[alias]))
                del node._extends[filename]

        self._update_format()
        LOGGER.info("\n%s", self._config.render())
        return self._config_flat

    def parse_repository(self, repofilename: str) -> repository.Repository:
        """
        Parse the given repository file.

        Executes the 'prepare' function to populate the repository
        structure.
        """
        repo = repository.Repository.parse_repository(repofilename)
        if repo.name in [_cw.name for _cw in self.children]:
            raise LbuildException("Repository name '{}' is ambiguous. Name must be unique."
                                  .format(repo.name))
        repo.parent = self

        return repo

    def merge_repository_options(self):
        # only deal with repo options that contain one `:`
        resolver = self.option_resolver
        for name, value in {n: v for n, v in self.config.options.items()
                            if n.count(":") == 1}.items():
            try:
                resolver[name].value = value
            except LbuildException as error:
                raise LbuildException("Failed to merge repository options!\n{}".format(error))

    def prepare_repositories(self):
        undefined = self._undefined_repo_options()
        if undefined:
            raise LbuildException("Unknown values for options '{}'. Please provide a value in the "
                                  "configuration file or on the command line."
                                  .format("', '".join(undefined)))
        modules = []
        for repo in self._findall(BaseNode.Type.REPOSITORY):
            modules.extend(repo.prepare())

        modules = lbuild.module.build_modules(modules)
        if not modules:
            raise LbuildBuildException("No module found with the selected repository options!")

        self._resolve_dependencies(ignore_failure=True)
        self._update_format()
        return modules

    def merge_module_options(self):
        # only deal with repo options that contain one `:`
        resolver = self.option_resolver
        for name, value in {n: v for n, v in self.config.options.items()
                            if n.count(":") > 1}.items():
            try:
                resolver[name].value = value
            except LbuildException as error:
                raise LbuildException("Failed to merge module options!\n{}".format(error))

    def resolve_dependencies(self, requested_modules, depth=sys.maxsize):
        # map the dependency names to the node objects
        self._resolve_dependencies()
        return self._filter_dependencies(requested_modules, depth)

    def _filter_dependencies(self, requested_modules, depth=sys.maxsize):
        """
        Resolve dependencies by adding missing modules.

        Args:
            modules (list): All modules available in the system
            requested_modules (list): Selected modules. Sub-set of all modules.
            depth (int): Maximum depth of dependencies. If not set to maximum
                not all dependencies might be resolved.

        Returns:
            list: Required modules for the given list of modules.
        """
        selected_modules = requested_modules.copy()

        LOGGER.info("Selected modules: %s",
                    ", ".join(sorted([module.fullname for module in selected_modules])))

        current = selected_modules
        while depth > 0:
            additional = []
            for module in current:
                for dependency in module.dependencies:
                    if dependency not in selected_modules and dependency not in additional:
                        LOGGER.debug("Add dependency: %s", dependency.fullname)
                        additional.append(dependency)
            if not additional:
                # Abort if no new dependencies are being found
                break
            selected_modules.extend(additional)
            current = additional
            additional = []
            depth -= 1

        # disable all non-selected modules
        for module in self.all_modules():
            if module._available:
                module._selected = module in selected_modules

        self._update()

        return selected_modules

    def find_module(self, name):
        return self.module_resolver[name]

    def find_option(self, name):
        return self.option_resolver[name]

    def find_modules(self, queries):
        return self.find_any(queries, self.Type.MODULE)

    def find_all(self, queries):
        return [node for q in utils.listify(queries) for node in self.find_any(q)]

    def find_any(self, queries, types=None):
        nodes = set()
        for query in utils.listify(queries):
            result = self._resolve_partial(query, None)
            if result is None:
                raise LbuildException("Cannot resolve '{}'".format(query))
            nodes |= set(result)
        if types:
            types = utils.listify(types)
            nodes = [n for n in nodes if any(n.type == t for t in types)]
        return list(nodes)

    @staticmethod
    def _undefined_options(options):
        undefined = []
        for fullname, option in options.items():
            if option.value is None:
                undefined.append(fullname)
        return undefined

    def _undefined_repo_options(self):
        return self._undefined_options(self.repo_options)

    @staticmethod
    def validate_modules(build_modules):
        Parser.build_modules(build_modules, None)

    @staticmethod
    def build_modules(build_modules, buildlog):
        if not build_modules:
            raise LbuildException("No modules selected, aborting!")

        groups = collections.defaultdict(list)
        for node in build_modules + list(set(m.repository for m in build_modules)):
            env = lbuild.environment.Environment(node, buildlog)
            groups[node.depth].append(Runner(node, env))

        exceptions = []
        # Enforce that the submodules are always build before their
        # parent modules.
        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                try:
                    runner.validate()
                except lbuild.exception.LbuildValidateException as error:
                    exceptions.append(error)

        if exceptions:
            raise lbuild.exception.LbuildAggregateException(exceptions)

        # Cannot build with these settings, just pre_build
        if buildlog is None:
            return

        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                runner.build()

        for index in sorted(groups, reverse=True):
            group = groups[index]
            random.shuffle(group)

            for runner in group:
                runner.post_build(buildlog)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2015-2018, Fabian Greif
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import re
import sys
import argparse
import traceback

import lbuild.logger
import lbuild.vcs.common
from lbuild.format import format_option_short_description

from lbuild.api import Builder

__version__ = '1.9.1'


class InitAction:

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "init",
            help="Load remote repositories into the cache folder.")
        parser.set_defaults(execute_action=self.perform)

    @staticmethod
    def perform(_, builder):
        lbuild.vcs.common.initialize(builder.config)
        return ""


class UpdateAction:

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "update",
            help="Update the content of remote repositories in the cache folder.")
        parser.set_defaults(execute_action=self.perform)

    @staticmethod
    def perform(_, builder):
        lbuild.vcs.common.update(builder.config)
        return ""


class ManipulationActionBase:
    """
    Base class for actions that interact directly with the parser repositories.

    All subclasses must implement a `perform` function.
    """
    # pylint: disable=too-few-public-methods

    def load_repositories(self, args, builder):
        builder.load(args.repositories)

        # Implemented by the subclasses, therefore not known to pylint
        # pylint: disable=no-member
        return self.perform(args, builder)


class DiscoverAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "discover",
            help="Render the available repository tree with modules and options. "
                 "You may need to provide options to see the entire tree!")
        parser.add_argument(
            dest="names",
            type=str,
            nargs="*",
            help="Select a specific repository, module or option.")
        parser.add_argument(
            "-n",
            "--name",
            dest="names_explicit",
            type=str,
            action="append",
            default=[],
            help="Select a specific repository, module or option.")
        parser.add_argument(
            "--values",
            dest="values",
            action="store_true",
            default=False,
            help="Display option values, instead of description")
        parser.add_argument(
            "-t",
            "--tree",
            action="store_true",
            default=False,
            dest="show_subtree",
            help="Show module tree instead of description.")
        parser.add_argument(
            "--developer",
            action="store_true",
            default=False,
            dest="show_developer_view",
            help="Show developer nodes in tree view and descriptions.")
        parser.set_defaults(execute_action=self.load_repositories)

    @staticmethod
    def perform(args, builder):
        if args.show_developer_view:
            lbuild.format.SHOW_NODES.update(lbuild.node.BaseNode.Type)
        names = args.names + args.names_explicit
        if names:
            ostream = []
            for node in builder.parser.find_all(names):
                if args.show_subtree:
                    ostream.append(node.render())
                else:
                    if args.values and node.type == node.Type.OPTION:
                        ostream.extend(node.values)
                    else:
                        ostream.append(node.description)
            if not args.values:
                return "\n\n\n\n".join(ostream)
            return "\n".join(ostream)

        return builder.parser.render()


class DiscoverOptionsAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "discover-options",
            help="Display all known option names, current values, allowed inputs and "
                 "short descriptions.")
        parser.add_argument(
            dest="names",
            type=str,
            nargs="*",
            help="Select a specific option.")
        parser.add_argument(
            "-n",
            "--name",
            dest="names_explicit",
            type=str,
            action="append",
            default=[],
            help="Select a specific option.")
        parser.set_defaults(execute_action=self.load_repositories)

    @staticmethod
    def perform(args, builder):
        names = args.names + args.names_explicit
        names = names if names else ["*", ":**"]
        nodes = builder.parser.find_any(names, (builder.parser.Type.MODULE,
                                                builder.parser.Type.REPOSITORY))
        options = [o for n in nodes for o in n.options]

        ostream = []
        for option in sorted(options, key=lambda n: (n.depth, n.fullname)):
            ostream.append(format_option_short_description(option))
            if option.short_description:
                ostream.append("")
                ostream.append(textwrap.indent(option.short_description, "  "))
            ostream.append("")

        return "\n".join(ostream)


class SearchAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "search",
            help="Search the descriptions of the repository tree and render the results "
                 "as a partial tree and the matching lines of the descriptions.")
        parser.add_argument(
            dest="queries",
            type=str,
            nargs="+",
            help="Regex-patters to be matched. Patterns are ORed together!")
        parser.set_defaults(execute_action=self.load_repositories)

    @staticmethod
    def perform(args, builder):
        nodes = builder.parser.find_any(["*", ":**"])
        nodes.sort(key=lambda n: n.fullname)
        show_nodes = lbuild.format.SHOW_NODES
        plain = lbuild.format.PLAIN

        search = "|".join(args.queries)
        fnodes = []
        ostream = ""
        lbuild.format.PLAIN = True

        for node in nodes:
            lbuild.format.SHOW_NODES = {node.type}
            description = node.description.splitlines()
            olines = []

            for line, descr in enumerate(description):
                if re.search(search, descr, flags=re.I):
                    olines.append("{:>3}  ".format(line-1) + descr)

            if olines:
                fnodes.append(node)
                ostream += "\n\n\n" + description[0]
                if olines[0].startswith(" -1  >> "):
                    olines = olines[1:]
                if olines:
                    ostream += "\n\n" + "\n".join(olines)

        tnodes = {a  for n in fnodes  for a in n.ancestors} | set(fnodes)
        ostream = builder.parser.render(lambda n: n.type in show_nodes and n in tnodes) + ostream

        if not plain:
            lbuild.format.PLAIN = False
            from lbuild.format import ansi_escape as c
            replace = r"{}\g<1>{}".format(str(c("bold")), str(c("no_bold")))
            ostream = re.sub(r"({})".format(search), replace, ostream, flags=re.I)
            replace = r">> {}\g<1>{}  [".format(str(c("bold")), str(c("no_bold")))
            ostream = re.sub(r">> (.*?)  \[", replace, ostream, flags=re.I)

        return ostream


class ValidateAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "validate",
            help="Validate the library configuration and data inputs with the given options.")
        parser.add_argument(
            "-m",
            "--module",
            dest="modules",
            type=str,
            action="append",
            default=[],
            help="Select a specific module.")
        parser.set_defaults(execute_action=self.load_repositories)

    @staticmethod
    def perform(args, builder):
        builder.validate(args.modules)
        return "Library configuration valid."


class BuildAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "build",
            help="Generate the library source code blob with the given options.")
        parser.add_argument(
            "-m",
            "--module",
            dest="modules",
            type=str,
            action="append",
            default=[],
            help="Select a specific module.")
        parser.add_argument(
            "--simulate",
            dest="simulate",
            action="store_true",
            default=False,
            help="Build, but do not write any files. Prints out all generated file names.")
        parser.add_argument(
            "--no-log",
            dest="buildlog",
            action="store_false",
            default=True,
            help="Do not create a build log. This log contains all files being "
                 "generated, their source files and the module which generated "
                 "the file.")
        parser.set_defaults(execute_action=self.load_repositories)

    @staticmethod
    def perform(args, builder):
        buildlog = builder.build(args.path, args.modules, simulate=args.simulate)

        if args.simulate:
            ostream = []
            for operation in buildlog.operations:
                ostream.append(operation.local_filename_out())
            return "\n".join(sorted(ostream))

        if args.buildlog:
            configfilename = args.config
            logfilename = configfilename + ".log"
            buildlog.log_unsafe("lbuild", "buildlog.xml.in", logfilename)
            with open(logfilename, "wb") as logfile:
                logfile.write(buildlog.to_xml(to_string=True, path=os.getcwd()))

        return ""


class CleanAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "clean",
            help="Remove previously generated files.")
        parser.add_argument(
            "--buildlog",
            dest="buildlog",
            default="project.xml.log",
            help="Use the given buildlog to identify the files to remove.")
        parser.set_defaults(execute_action=self.perform)

    @staticmethod
    def perform(args, builder):
        ostream = []
        if os.path.exists(args.buildlog):
            with open(args.buildlog, "rb") as logfile:
                buildlog = lbuild.buildlog.BuildLog.from_xml(logfile.read(), path=os.getcwd())
        else:
            builder.load(args.repositories)
            buildlog = builder.build(args.path, simulate=True)

        dirs = set()
        filenames = [op.local_filename_out() for op in buildlog.operations]
        for filename in sorted(filenames):
            ostream.append("Removing " + filename)
            dirs.add(os.path.dirname(filename))
            try:
                os.remove(filename)
            except OSError:
                pass

        dirs = sorted(list(dirs), key=lambda d: d.count("/"), reverse=True)
        for directory in dirs:
            try:
                os.removedirs(directory)
            except OSError:
                pass

        return "\n".join(ostream)


class DependenciesAction(ManipulationActionBase):

    def register(self, argument_parser):
        parser = argument_parser.add_parser(
            "dependencies",
            help="Generate a grahpviz representation of the module dependencies.")
        parser.add_argument(
            "-m",
            "--module",
            dest="modules",
            type=str,
            action="append",
            default=[],
            help="Select specific modules.")
        parser.add_argument(
            "-n",
            "--depth",
            dest="depth",
            type=int,
            default=sys.maxsize,
            help="Only show dependencies up to a specific depth. Only valid if "
                 "specific modules are selected, otherwise all modules are printed "
                 "anyways.")
        parser.set_defaults(execute_action=self.load_repositories)

    @staticmethod
    def perform(args, builder):
        selected_modules = args.modules + builder.parser.config.modules
        if not selected_modules:
            selected_modules = [":**"]
        dot_file = lbuild.builder.dependency.graphviz(builder,
                                                      selected_modules,
                                                      args.depth)
        return dot_file


def prepare_argument_parser():
    """
    Set up the argument parser for the different commands.

    Return:
    Configured ArgumentParser object.
    """
    argument_parser = argparse.ArgumentParser(
        description='Build source code libraries from modules.')
    argument_parser.add_argument(
        '-r',
        '--repository',
        metavar="REPO",
        dest='repositories',
        action='append',
        default=[],
        help="Repository file(s) which should be available for the current library. "
             "The loading of repository files from a VCS is only supported through "
             "the library configuration file.")
    argument_parser.add_argument(
        '-c',
        '--config',
        dest='config',
        default='project.xml',
        help="Project configuration file. "
             "Specifies the required repositories, modules and options "
             "(default: '%(default)s').")
    argument_parser.add_argument(
        '-p',
        '--path',
        dest='path',
        default='.',
        help="Path in which the library will be generated (default: '%(default)s').")
    argument_parser.add_argument(
        '-D',
        '--option',
        metavar='OPTION',
        dest='options',
        action='append',
        type=str,
        default=[],
        help="Additional options. Options given here will be merged with options "
             "from the configuration file and will overwrite the configuration "
             "file definitions.")
    argument_parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        dest='verbose')
    argument_parser.add_argument(
        "--plain",
        dest="plain",
        action="store_true",
        default=not sys.stdout.isatty(),
        help="Disable styled output, only output plain ASCII.")
    argument_parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
        help="Print the lbuild version number and exit.")

    subparsers = argument_parser.add_subparsers(
        title="Actions",
        dest="action")

    actions = [
        DiscoverAction(),
        DiscoverOptionsAction(),
        SearchAction(),

        ValidateAction(),
        BuildAction(),
        CleanAction(),

        InitAction(),
        UpdateAction(),
        DependenciesAction(),
    ]
    for action in actions:
        action.register(subparsers)

    return argument_parser


def run(args):
    lbuild.logger.configure_logger(args.verbose)
    lbuild.facade.VERBOSE_DEPRECATION = args.verbose
    lbuild.format.PLAIN = args.plain

    try:
        command = args.execute_action
    except AttributeError:
        raise lbuild.exception.LbuildArgumentException("No command specified")

    builder = Builder(config=args.config, options=args.options)
    return command(args, builder)


def main():
    """
    Main entry point of lbuild.
    """
    try:
        argument_parser = prepare_argument_parser()

        commandline_arguments = sys.argv[1:]
        args = argument_parser.parse_args(commandline_arguments)
        lbuild.logger.configure_logger(args.verbose)

        output = run(args)
        print(output)
    except lbuild.exception.LbuildAggregateException as aggregate:
        for error in aggregate.exceptions:
            sys.stderr.write('\nERROR: %s\n' % error)
        sys.exit(2)
    except lbuild.exception.LbuildForwardException as error:
        sys.stderr.write("\nERROR in '{}'\n".format(error.module))
        traceback.print_exception(type(error.exception),
                                  error.exception,
                                  error.exception.__traceback__,
                                  limit=-1)
        sys.exit(4)
    except lbuild.exception.LbuildArgumentException as error:
        argument_parser.print_help()
        print(error)
        sys.exit(2)
    except lbuild.exception.LbuildTemplateException as error:
        sys.stderr.write('\nERROR: %s\n' % error)
        traceback.print_exc()
        sys.exit(3)
    except lbuild.exception.LbuildException as error:
        sys.stderr.write('\nERROR: %s\n' % error)
        if args.verbose >= 2:
            traceback.print_exc()
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2018, Niklas Hauser
# All Rights Reserved.
#
# The file is part of the lbuild project and is released under the
# 2-clause BSD license. See the file `LICENSE.txt` for the full license
# governing this code.

import os
import sys
import shutil
import anytree
import colorful

import lbuild.node
import lbuild.filter

PLAIN = not sys.stdout.isatty()
WIDTH = shutil.get_terminal_size((100, 0)).columns
SHOW_NODES = {
    lbuild.node.BaseNode.Type.REPOSITORY,
    lbuild.node.BaseNode.Type.MODULE,
    lbuild.node.BaseNode.Type.OPTION,
    lbuild.node.BaseNode.Type.CONFIG,
}

COLOR_SCHEME = {
    "parser": None,
    "repository": None,
    "option": None,
    "query": None,
    "config": None,
    "collector": None,
    "module": None,
    "description": None,
    "short_description": None,
    "error": colorful.red,  # pylint: disable=no-member
}


def ansi_escape(obj=None):
    if PLAIN:
        # The terminal does not support color
        return ""

    name = obj
    if isinstance(obj, lbuild.node.BaseNode):
        name = obj.type.name.lower()

    col = COLOR_SCHEME.get(name, "nope")
    if col == "nope":
        try:
            col = getattr(colorful, name)
        except AttributeError:
            col = None

    return str(col) if col is not None else ""


class ColorWrapper:

    def __init__(self, string=None):
        self._string = ""
        self._content = []

        if isinstance(string, ColorWrapper):
            self._string = string._string
            self._content = string._content.copy()
        elif string is not None:
            self._string = string
            self._content = [("", self._string)]

    def wrap(self, name):
        style = ansi_escape(name)
        if style is not None:
            if name in ["underlined", "bold"]:
                close = ansi_escape("no_" + name)
            else:
                close = ansi_escape("close_fg_color")
            self._content = [(style, "")] + self._content + [(close, "")]
        return self

    def join(self, strings):
        result = ColorWrapper()
        for index, string in enumerate(strings):
            result += ColorWrapper(string)
            if index < len(strings) - 1:
                result += self
        return result

    def limit(self, offset=0):
        if len(self) <= (WIDTH - offset):
            return str(self)

        mark = " ..."
        width = WIDTH - offset - len(mark)
        if width < 0:
            width = 0
        style_str = ""
        raw_string = ""
        for style, content in self._content:
            style_str += style
            if len(raw_string + content) >= width:
                style_str += content[:width - len(raw_string)]
                return style_str + ansi_escape("reset") + mark
            raw_string += content
            style_str += content
        return style_str

    def __add__(self, other):
        color = ColorWrapper()
        if isinstance(other, ColorWrapper):
            color._content = self._content + other._content
            color._string = self._string + other._string
        else:
            color._content = self._content
            color._string = other
        return color

    def __len__(self):
        return len(self._string)

    def __iter__(self):
        return iter(self._string)

    def __contains__(self, string):
        return string in self._string

    def __getattr__(self, name):
        color = ansi_escape(name)
        if color is not None:
            self._content += [(color, "")]
        return self

    def __eq__(self, other):
        return self._string == other._string

    def __ne__(self, other):
        return self._string != other._string

    def __str__(self):
        string = ""
        for style, content in self._content:
            string += style + content
        return string + ansi_escape("reset")


_cw = ColorWrapper


def format_option_name(node, fullname=True):
    line = _cw(node.fullname if fullname else node.name).wrap(node).wrap("bold")
    if not node.is_default():
        line.wrap("underlined")
    return line


def format_option_value(node, single_line=True):
    offset = -1
    if node.value is None:
        return (_cw("REQUIRED").wrap("error").wrap("bold"), offset, False)
    value = node.format_value()
    if single_line:
        value = value.replace("\n", " ")
    return (_cw(value).wrap("bold"), offset, True)


def format_option_values(node, offset=0, single_line=True):
    values = node.format_values()
    if not single_line:
        if (offset + len(values)) > WIDTH:
            values = lbuild.filter.indent(lbuild.filter.wordwrap(values._string, WIDTH - offset),
                                          offset)
    return _cw(values).wrap("bold")


def format_option_value_description(node, offset=0, single_line=None):
    value = format_option_value(node, bool(single_line))
    offset = offset + len(value[0]) if value[1] < 0 else value[1]
    single_line = value[2] if single_line is None else single_line
    values = format_option_values(node, offset + 5, single_line)
    return value[0] + _cw(" in [") + values + _cw("]")


def format_option_short_description(node):
    line = format_option_name(node) + _cw(" = ")
    line += format_option_value_description(node, offset=len(line))
    if "REQUIRED" not in line:
        line = line.limit()
    return str(line)


def format_description(node, description):
    type_description = "  [{}]".format(node.class_name)
    output = [_cw(">> ") + _cw(node.description_name).wrap(node).wrap("bold") + _cw(type_description)]
    if description:
        description = description.strip()
        if len(description):
            output += [_cw(""), _cw(description)]

    if node.type == node.Type.OPTION:
        value = format_option_value(node, single_line=False)[0]
        values = format_option_values(node, offset=9, single_line=False)
        output += [_cw(""), _cw("Value: ") + value, _cw("Inputs: [") + values + _cw("]")]
    elif node.type == node.Type.COLLECTOR:
        values = format_option_values(node, offset=9, single_line=False)
        output += [_cw(""), _cw("Inputs: [") + values + _cw("]")]

    children = []
    # Print every node except the submodule, due to obvious recursion
    for ntype in (SHOW_NODES - {lbuild.node.BaseNode.Type.MODULE}):
        children += node._findall(ntype, depth=2)

    for child in sorted(children, key=lambda c: (c.type, c.name)):
        output.extend([_cw("\n"), _cw(">>") + _cw(child.description)])

    return "\n".join(map(str, output))


def format_short_description(_, description):
    lines = description.strip().splitlines() + [""]
    return lines[0].strip().rstrip(".,:;!?")


def format_node(node, _):
    name = _cw(node.name)
    class_name = _cw(node.class_name)
    if node._type == node.Type.REPOSITORY:
        name = _cw(node.name + " @ " + os.path.relpath(node._filepath, os.getcwd()))
    elif node._type == node.Type.OPTION:
        name = format_option_name(node, fullname=False)
    elif node._type in [node.Type.MODULE, node.Type.CONFIG]:
        name = _cw(node.fullname).wrap(node)
    elif node._type in [node.Type.QUERY, node.Type.COLLECTOR]:
        class_name = class_name.wrap("underlined")
        name = name.wrap("bold")

    descr = (class_name + _cw("(") + name + _cw(")")).wrap(node)

    offset = node.depth * 4
    if node._type == node.Type.OPTION:
        descr += _cw(" = ")
        descr += format_option_value_description(node, offset=offset + len(descr), single_line=True)
    elif node._type == node.Type.COLLECTOR:
        descr += _cw(" in [")
        descr += format_option_values(node, offset=offset + len(descr), single_line=True)
        descr += _cw("]")

    descr += _cw("   " + node.short_description)

    return descr.limit(offset)


def format_node_tree(node, filterfunc=None):
    if filterfunc is None:
        filterfunc = lambda n: n.type in SHOW_NODES

    class Renderer(anytree.RenderTree):
        def __init__(self, node):
            anytree.RenderTree.__init__(self, node,
                                        style=anytree.ContRoundStyle(),
                                        childiter=self.childiter)

        @staticmethod
        def childiter(nodes):
            nodes = [n for n in nodes if filterfunc(n)]
            return sorted(nodes, key=lambda node: (node._type, node.name))

        def __str__(self):
            lines = []
            for pre, _, node in self:
                lines.append(pre + format_node(node, pre))
            return "\n".join(lines)

    return str(Renderer(node))